import re
import sys
from abc import ABC
from functools import cached_property, lru_cache
from os.path import expandvars, expanduser
from pathlib import Path
from time import sleep
from typing import List, Tuple, Union, Optional, Type
from urllib.parse import urljoin

from bs4 import BeautifulSoup, Tag, ResultSet
from requests import get
from requests_cache import CachedSession

from enmet.common import CachedInstance

__all__ = ["set_session_cache"]

_METALLUM_URL = "https://www.metal-archives.com"
# Without correct user-agent there are 4xx responses
_USER_AGENT = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko)" \
              "Chrome/102.0.5005.167 Safari/537.36"


def _split_by_sep(data: str) -> List[str]:
    """Split different text list (genres, lyrical themes etc.) into separate parts."""
    return re.split(r"\s*[,;]\s*", data.strip())


class _Page(ABC):
    # Path part of URL
    RESOURCE = None


class _SearchResultsPage(_Page, ABC):
    def __init__(self, params):
        self.params = params

    def _fetch_search_result(self):
        """Note: iDisplayLength does not work"""
        params = self.params
        data = get(urljoin(_METALLUM_URL, self.RESOURCE),
                   params=params,
                   headers={"User-Agent": _USER_AGENT}
                   ).json()
        if len(data["aaData"]) == 0:
            return []
        records = data["aaData"]
        while len(records) < data["iTotalRecords"]:
            self.params["iDisplayStart"] = len(records)
            data = get(urljoin(_METALLUM_URL, self.RESOURCE),
                       params=params,
                       headers={"User-Agent": _USER_AGENT}
                       ).json()
            records.extend(data["aaData"])
        return records


class BandSearchPage(_SearchResultsPage):
    RESOURCE = "search/ajax-advanced/searching/bands"

    @cached_property
    def bands(self) -> List[Tuple[str, ...]]:
        records = self._fetch_search_result()
        result = []
        for item in records:
            bs = BeautifulSoup(item[0], features="html.parser")
            band_link, band = bs.select_one("a")["href"], bs.select_one("a").text
            genres = item[1]
            country = item[2]  # Location if searched with single country
            formed = None
            if len(item) == 4:  # May be not present
                formed = item[3]
            result.append((band_link, band, genres, country, formed))
        return result


class AlbumSearchPage(_SearchResultsPage):
    RESOURCE = "search/ajax-advanced/searching/albums/"

    @cached_property
    def albums(self) -> List[Tuple[str, ...]]:
        records = self._fetch_search_result()
        result = []
        for item in records:
            bs = BeautifulSoup(item[0], features="html.parser")
            band_link, band = bs.select_one("a")["href"], bs.select_one("a").text
            bs = BeautifulSoup(item[1], features="html.parser")
            album_link, album = bs.select_one("a")["href"], bs.select_one("a").text
            release_date = item[3][:item[3].find("<")]
            result.append((album_link, album, band_link, band, release_date))
        return result


class _CachedSite:
    """Virtual The Metal Archives site. Descriptor getting data from Metal Archives site for supported classes."""
    _CACHE_PATH = Path(expandvars("%LOCALAPPDATA%") if sys.platform == "win32" else expanduser("~")) / ".enmet"
    _CACHE_NAME = "enmet_data"
    # 1/QUERY_RATE second is the delay after each non-cached _DataPage read from the site
    QUERY_RATE = 3
    # Get a bit more performance maybe by saving time on reading from database and converting to BeautifulSoup objects?
    _BS_CACHE_SIZE = 100

    def __init__(self):
        self._session = None

    def set_session(self, **kwargs) -> CachedSession:
        """Factory method for CachedSession with delay hook."""
        session = CachedSession(
            **({"cache_name": str(self._CACHE_PATH / self._CACHE_NAME), "backend": "sqlite"} | kwargs))
        session.hooks['response'].append(
            lambda r, *args, **kwargs: None if not getattr(r, "from_cache", False) and sleep(
                1 / _CachedSite.QUERY_RATE) else None)
        self._session = session
        return session

    @lru_cache(maxsize=_BS_CACHE_SIZE)
    def _cached_get(self, resource: str) -> BeautifulSoup:
        """Get page from Metal Archives with caching."""
        response = self._session.get(urljoin(_METALLUM_URL, resource),
                                     headers={"User-Agent": _USER_AGENT, 'Accept-Encoding': 'gzip'}
                                     )
        response.raise_for_status()
        return BeautifulSoup(response.content, features="html.parser")

    def __get__(self, instance, owner) -> Union[BeautifulSoup, "_CachedSite"]:
        """Method returning page or search results from Metal Archives"""
        if instance is None:
            return self
        if self._session is None:  # Lazy session creation to enable setting cache before it is accessed.
            self._CACHE_PATH.mkdir(parents=True, exist_ok=True)
            self.set_session()
        resource = instance.RESOURCE.format(instance.id)
        return self._cached_get(resource)


class _DataPage(_Page, CachedInstance, ABC):

    enmet = _CachedSite()

    def __init__(self, id_: str):
        self.id = str(id_)

    def _get_header_item(self, name: str) -> Optional[Tag]:
        elem = self.enmet.find("dt", string=name)
        return elem.find_next_sibling() if elem else None

    @staticmethod
    def set_session_cache(**kwargs) -> CachedSession:
        return _DataPage.enmet.set_session(**kwargs)


class DiscographyPage(_DataPage):
    RESOURCE = "band/discography/id/{}/tab/all"

    @cached_property
    def albums(self) -> List[List[Optional[str]]]:
        result = []
        for elem in self.enmet.select(".discog tbody tr"):
            # Album URL, Name
            result.append([elem.select_one("td:nth-child(1) a")["href"], elem.select_one("td:nth-child(1) a").text])
            # Type
            result[-1].append(elem.select_one("td:nth-child(2)").text)
            # Year
            result[-1].append(elem.select_one("td:nth-child(3)").text)
            # Reviews URL, Reviews
            if (e := elem.select_one("td:nth-child(4)")).text.strip():
                result[-1].extend([e.select_one("a")["href"], e.select_one("a").text])
            else:
                result[-1].extend([None, None])
        return result


class BandPage(_DataPage):
    RESOURCE = "bands/_/{}"

    @cached_property
    def name(self):
        return self.enmet.select_one(".band_name a").text

    @cached_property
    def country(self) -> Optional[str]:
        return elem.text if (elem := self._get_header_item("Country of origin:")) else None

    @cached_property
    def location(self) -> Optional[str]:
        return elem.text if (elem := self._get_header_item("Location:")) else None

    @cached_property
    def status(self) -> Optional[str]:
        return elem.text if (elem := self._get_header_item("Status:")) else None

    @cached_property
    def formed_in(self) -> Optional[str]:
        return elem.text if (elem := self._get_header_item("Formed in:")) else None

    @cached_property
    def years_active(self):
        return _split_by_sep(self._get_header_item("Years active:").text.strip())

    @cached_property
    def genres(self) -> List[str]:
        return _split_by_sep(self._get_header_item("Genre:").text.strip())

    @cached_property
    def lyrical_themes(self) -> List[str]:
        return _split_by_sep(self._get_header_item("Lyrical themes:").text.strip())

    @cached_property
    def current_label(self):
        return elem.text if (elem := self._get_header_item("Current label:")) else None

    @cached_property
    def last_label(self):
        return elem.text if (elem := self._get_header_item("Last label:")) else None

    @staticmethod
    def _get_members_list(rows: ResultSet[Tag]) -> List[List[Optional[str]]]:
        result = []
        for elem in rows:
            # Artist URL, Artist
            result.append([elem.select_one("a")["href"], elem.select_one("a").text])
            # Role
            result[-1].append(elem.select_one("td:nth-child(2)").text.replace("\n", " ").replace("\xa0", " ").strip())
        return result

    @cached_property
    def lineup(self) -> List[List[Optional[str]]]:
        rows = self.enmet.select("#band_tab_members_current tr.lineupRow")
        return self._get_members_list(rows)

    @cached_property
    def past_members(self) -> List[List[Optional[str]]]:
        rows = self.enmet.select("#band_tab_members_past tr.lineupRow")
        return self._get_members_list(rows)

    @cached_property
    def live_musicians(self) -> List[List[Optional[str]]]:
        rows = self.enmet.select("#band_tab_members_live tr.lineupRow")
        return self._get_members_list(rows)

    @cached_property
    def info(self) -> str:
        if self.enmet.select_one(".band_comment a.btn_read_more"):
            return _BandInfoPage(self.id).info.strip()
        else:
            return " ".join(e.text.strip() for e in self.enmet.select_one(".band_comment").contents).strip()

    @cached_property
    def last_modified(self) -> str:
        return self.enmet.find("td", string=re.compile("Last modified on")).text


class _BandInfoPage(_DataPage):
    RESOURCE = "band/read-more/id/{}"

    @cached_property
    def info(self) -> str:
        return self.enmet.text


class BandRecommendationsPage(_DataPage):
    RESOURCE = "band/ajax-recommendations/id/{}/showMoreSimilar/1"

    @cached_property
    def similar_artists(self) -> List[List[str]]:
        rows = self.enmet.select("#artist_list tr:not(:last-child)")
        results = []
        if len(rows) and rows[0].text.startswith("No similar artist"):
            return results
        for row in rows:
            data = row.select("td")
            results.append([data[0].select_one("a")["href"], data[0].text])  # Band URL, band name
            results[-1].append(data[1].text)  # Country
            results[-1].append(data[2].text)  # Genre
            results[-1].append(data[3].text)  # Score
        return results


class AlbumPage(_DataPage):
    RESOURCE = "albums/_/_/{}"

    @cached_property
    def name(self):
        return self.enmet.select_one(".album_name a").text

    @cached_property
    def bands(self) -> List[Tuple[str, str]]:
        """List of album bands, more than 1 for splits, cooperations etc."""
        bands = []
        elems = self.enmet.select("#album_info .band_name a")
        for b in elems:
            bands.append((b["href"], b.text))
        return bands

    @cached_property
    def type(self):
        return self._get_header_item("Type:").text

    @cached_property
    def release_date(self):
        return self._get_header_item("Release date:").text

    @cached_property
    def catalog_id(self):
        return self._get_header_item("Catalog ID:").text

    @cached_property
    def label(self):
        return self._get_header_item("Label:").text

    @cached_property
    def format(self):
        return self._get_header_item("Format:").text

    @cached_property
    def reviews(self) -> Tuple[Optional[str], str]:
        elem = self._get_header_item("Reviews:")
        if elem.select_one("a"):
            return elem.select_one("a")["href"], elem.text
        else:
            return None, elem.text

    @cached_property
    def tracks(self) -> List[List[Union[str, Optional[bool]]]]:
        result = [[]]
        for elem in self.enmet.select_one("#album_tabs_tracklist").select("tr.even,tr.odd,.discRow"):
            if "discRow" in elem["class"]:
                if len(result[0]) != 0:  # Another disc
                    result.append([])
                continue
            # Id
            result[-1].append([elem.select_one("td:nth-of-type(1) a")["name"]])
            # Number
            number = elem.select_one("td:nth-of-type(1)").text
            result[-1][-1].append(int(number[:number.index(".")]))
            # Name
            result[-1][-1].append(elem.select_one("td:nth-of-type(2)").text.strip())
            # Time
            result[-1][-1].append(elem.select_one("td:nth-of-type(3)").text)
            # Lyrics status
            lyrics = elem.select_one("td:nth-of-type(4)")
            if lyrics.select_one("a"):  # Has lyrics
                result[-1][-1].append(True)
            elif lyrics.select_one("em"):  # Marked as instrumental
                result[-1][-1].append(False)
            else:
                result[-1][-1].append(None)
        return result

    @cached_property
    def disc_names(self) -> List[Optional[str]]:
        return [e.text for e in self.enmet.select(".discRow td")] or [None]

    @cached_property
    def total_times(self) -> List[Optional[str]]:
        return [e.text for e in self.enmet.select(".table_lyrics strong")] or [None]

    @cached_property
    def lineup(self):
        result = []
        for elem in self.enmet.select("#album_members_lineup tr.lineupRow"):
            # Artist URL, Artist
            result.append([elem.select_one("a")["href"], elem.select_one("a").text])
            # Role
            result[-1].append(elem.select_one("td:nth-child(2)").text.strip())
        return result


class ArtistPage(_DataPage):
    RESOURCE = "artists/_/{}"

    @cached_property
    def name(self):
        return self.enmet.select_one(".band_member_name").text

    @cached_property
    def real_full_name(self):
        return self._get_header_item("Real/full name:").text.strip()

    @cached_property
    def age(self) -> str:
        return self._get_header_item("Age:").text.strip()

    @cached_property
    def place_of_birth(self) -> str:
        return self._get_header_item("Place of birth:").text.strip()

    @cached_property
    def gender(self) -> str:
        return self._get_header_item("Gender:").text

    def _get_extended_section(self, caption: str, cls_data_source: Type[_DataPage]) -> Optional[str]:
        # This is a mess because the HTML for this section is a mess...
        top = self.enmet.select_one("#member_content .band_comment")
        if caption_elem := top.find("h2", string=caption):
            idx_caption = top.index(caption_elem)
            has_readme = False
            idx = 0
            for idx, elem in enumerate(top.contents[idx_caption+1:]):
                if not isinstance(elem, Tag):
                    continue
                elif elem.text == "Read more":
                    has_readme = True
                    break
                elif elem.name == "h2":
                    break
            else:
                idx += 1
            if has_readme:
                return getattr(cls_data_source(self.id), caption.lower()).strip()
            else:
                return " ".join([e.text.strip() for e in top.contents[idx_caption+1:idx_caption+1+idx]])
        else:
            return None

    @cached_property
    def biography(self) -> Optional[str]:
        return self._get_extended_section("Biography", _ArtistBiographyPage)

    @cached_property
    def trivia(self) -> Optional[str]:
        return self._get_extended_section("Trivia", _ArtistTriviaPage)


class _ArtistBiographyPage(_DataPage):
    RESOURCE = "artist/read-more/id/{}"

    @cached_property
    def biography(self) -> str:
        return self.enmet.text


class _ArtistTriviaPage(_DataPage):
    RESOURCE = "artist/read-more/id/{}/field/trivia"

    @cached_property
    def trivia(self) -> str:
        return self.enmet.text


class LyricsPage(_DataPage):
    RESOURCE = "release/ajax-view-lyrics/id/{}"

    # az lyrics
    # darklyrics
    # GENIUS
    @cached_property
    def lyrics(self):
        return self.enmet.get_text().strip()


class RandomBandPage:
    @cached_property
    def band(self) -> str:
        data = get(urljoin(_METALLUM_URL, "band/random"),
                   headers={"User-Agent": _USER_AGENT})
        return data.url


def set_session_cache(**kwargs) -> CachedSession:
    """Set cache for DataPages reads"""
    return _DataPage.set_session_cache(**kwargs)



from datetime import timedelta, datetime
from pathlib import Path
from types import SimpleNamespace
from unittest.mock import call, ANY

import pytest
from bs4 import BeautifulSoup

from enmet import set_session_cache, search_bands, Artist, PartialDate, Band, Countries, search_albums, ReleaseTypes, \
    Album, Track, EnmetEntity
from enmet.common import datestr_to_date
from enmet.pages import _CachedSite


@pytest.fixture(scope="session", autouse=True)
def temp_cache():
    Path("enmet_test_dummy.sqlite").unlink(missing_ok=True)
    set_session_cache(cache_name="enmet_test_dummy")
    yield


def test_band():
    band = search_bands(name="megadeth", strict=True, formed_from=1980)[0]
    assert repr(band) == "<Band: Megadeth (138)>"
    assert str(band) == "Megadeth (United States)"
    assert band.name == "Megadeth"
    assert str(band.country) == "United States"
    assert band.location == "Los Angeles, California"
    assert band.formed_in == 1983
    assert band.years_active == ["1983 (as Fallen Angels)", "1983-2002", "2004-present"]
    assert band.genres == ["Thrash Metal (early/later)", "Heavy Metal/Rock (mid)"]
    assert band.lyrical_themes == ["Politics", "War", "History", "Death", "Religion", "Society", "New World Order"]
    assert band.label == "Tradecraft"
    assert {a.id for a in band.lineup} == {"184", "2836", "3826", "1391"}
    assert repr(band.lineup[0]) == "<LineupArtist: Dave Mustaine (184)>"
    assert str(band.lineup[0]) == "Dave Mustaine"
    assert Artist(184) is band.lineup[0].artist
    assert band.lineup[0].artist.real_full_name == "David Scott Mustaine"
    assert str(band.lineup[0].artist) == "Dave Mustaine"
    assert all(x in dir(band.lineup[0]) for x in ["name_in_lineup", "band"])
    assert len(band.past_members) > 22
    assert len(band.live_musicians) > 5
    assert band.discography[0].release_date == PartialDate(year=1984, month="March", day=9)
    assert repr(band.discography[0].discs[0]) == "<Disc: None>"
    assert set(dir(band)) == {'country', 'discography', 'formed_in', 'genres', 'info', 'label', 'last_modified',
                              'lineup', 'live_musicians', 'location', 'lyrical_themes', 'name', 'past_members',
                              'similar_artists', 'status', 'years_active'}
    # similar_artists
    assert len(band.similar_artists) > 180
    assert band.similar_artists[0].score > 490
    assert band.similar_artists[0].similar_to is band
    assert band.similar_artists[0].name == "Metallica"
    assert "Metallica" in repr(band.similar_artists[0])
    assert set(dir(band.similar_artists[0])) == {'country', 'discography', 'formed_in', 'genres', 'info', 'label',
                                                 'last_modified', 'lineup', 'live_musicians', 'location',
                                                 'lyrical_themes', 'name', 'past_members', 'score', 'similar_artists',
                                                 'similar_to', 'status', 'years_active'}
    assert band.info.startswith("Pictured from left to right")
    assert band.last_modified >= datetime(2022, 10, 10, 15, 58, 54)


def test_band_no_formed_in_no_biography():
    # given
    b = search_bands(name="witches of moodus")[0]
    # then
    assert b.formed_in is None
    assert b.info.startswith("Compilation")
    # assert b.lineup[1].biography is None  # no Trivia or Biography section


def test_artist():
    # given
    a = Artist(184)
    # then
    assert "1961" in a.age
    assert a.place_of_birth == 'United States (La Mesa, California)'
    assert a.gender == "Male"
    assert a.biography.startswith("Mustaine was born in La Mesa")
    assert a.trivia.startswith("Dave performed alongside Dream Theater")
    assert set(dir(a)) == {'age', 'biography', 'gender', 'name', 'place_of_birth', 'real_full_name', 'trivia'}


def test_artist_two_extended_sections_first_no_read_more():
    # given
    a = Artist(107)
    # then
    assert a.biography.startswith("Adrian Smith is an English guitarist")


def test_artist_less_extras():
    # given
    a = Artist(14883)
    # then
    assert a.trivia.startswith("DiSanto was arrested")
    assert a.biography is None


def test_band_splitup():
    band = Band("72")
    assert band.label == "Nuclear Blast"


def test_search_bands_set_country(mocker):
    # given
    asp_mock = mocker.patch("enmet.search.BandSearchPage")
    # when
    search_bands(name="dummy", countries=[Countries.POLAND])
    # then
    assert asp_mock.mock_calls[0] == call(
        {'bandName': 'dummy', 'exactBandMatch': '', 'genre': '', 'yearCreationFrom': '', 'yearCreationTo': '',
         'country[]': ['PL']})


def test_search_bands_no_params():
    # when
    bands = search_bands()
    # then
    assert bands == []


def test_album():
    album = search_albums(name="Metal Heart", band="Accept")[0]
    assert {a.id for a in album.lineup} == {'21647', '21529', '21621', '21592', '21656'}
    assert album.discs[0].tracks[0].lyrics.startswith("It is 1999")
    assert album.type == ReleaseTypes.FULL
    assert album.catalog_id == "PL 70638"
    assert album.label == "RCA"
    assert album.format == '12" vinyl (33⅓ RPM)'
    assert album.reviews == ('https://www.metal-archives.com/reviews/Accept/Metal_Heart/826/', '\n12 reviews (avg. 78%)\n')
    assert album.total_time == timedelta(minutes=39, seconds=55)
    assert repr(album) == "<Album: Metal Heart (826)>"
    assert str(album) == "Metal Heart (1985)"
    assert album.discs[0].number == 1
    assert album.discs[0].tracks[0].band is Band("198")
    assert album.year == 1985
    assert dir(album) == ['bands', 'catalog_id', 'discs', 'format', 'label', 'lineup', 'name', 'release_date',
                          'reviews', 'total_time', 'type', 'year']
    assert set(dir(album.lineup[0])) == {'album', 'name', 'name_on_album', 'real_full_name', 'role', 'age', 'biography',
                                         'gender', 'place_of_birth', 'trivia'}
    assert dir(album.discs[0]) == ['name', 'number', 'total_time', 'tracks']
    assert dir(album.discs[0].tracks[0]) == ['band', 'lyrics', 'name', 'number', 'time']


def test_search_albums_with_years(mocker):
    # given
    asp_mock = mocker.patch("enmet.search.AlbumSearchPage")
    # when
    search_albums(name="dummy", year_from=1991, year_to=1992)
    # then
    assert asp_mock.mock_calls[0] == call(
        {'releaseTitle': 'dummy', 'exactReleaseMatch': '', 'bandName': '', 'exactBandMatch': '',
         'releaseYearFrom': 1991, 'releaseMonthFrom': '', 'releaseYearTo': 1992, 'releaseMonthTo': '', 'genre': '',
         'releaseType[]': []})


def test_search_albums_no_params():
    # when
    albums = search_albums()
    # then
    assert albums == []


def test_album_missing_values():
    # given
    album = Album("3509")
    # then
    assert album.name == "World War III"
    assert album.discs[0].tracks[2].name == "Vindicator"
    assert album.discs[0].tracks[2].time is None
    assert album.discs[0].tracks[2].lyrics is None
    assert album.discs[0].name is None


def test_album_sides_instrumental():
    album = Album("457889")
    assert len(album.discs[0].tracks) == 8
    assert album.discs[0].tracks[2].name == "The Ytse Jam"
    assert album.discs[0].tracks[2].lyrics is False


def test_album_named_discs():
    album = Album("534606")
    assert album.name == "Songs from the North I, II & III"
    assert [disc.name for disc in album.discs] == ["Gloom", "Beauty", "Despair"]


def test_album_split():
    album = Album("907847")
    assert {band.name for band in album.bands} == {"Vektor", "Cryptosis"}
    assert album.discs[0].tracks[0].band.name == "Vektor"
    assert album.discs[0].tracks[0].name == "Activate"
    assert album.discs[0].tracks[2].band.name == "Cryptosis"


def test_album_disc_and_sides():
    album = Album("78")
    assert len(album.discs) == 2
    assert len(album.discs[0].tracks) == 8
    assert len(album.discs[1].tracks) == 9


def test_no_search_results():
    # given
    result = search_albums(name="XYZ<>123")
    # then
    assert result == []


def test_big_search():
    # given
    result = search_bands(name="fire", strict=False)
    # then
    assert len(result) > 200


def test_partialdate_bad_month():
    with pytest.raises(ValueError):
        PartialDate(year=1999, month="dummy", day=1)


def test_partialdate_bad_day():
    with pytest.raises(ValueError):
        PartialDate(year=1999, month="September", day=99)


def test_partial_date_str_repr():
    # given
    pd = PartialDate(1999, "March", 7)
    # then
    assert str(pd) == "1999-03-07"
    assert repr(pd) == "<PartialDate: year=1999, month=3, day=7>"


@pytest.mark.parametrize("datestr, year, month, day",
                         [("March 1999", 1999, 3, None), ("2004", 2004, None, None)])
def test_datestr_to_date(datestr, year, month, day):
    # given
    result = datestr_to_date(datestr)
    # then
    assert result.year == year
    assert result.month == month
    assert result.day == day


def test_Track_no_band_for_track():
    # given
    b1, b2 = SimpleNamespace(name="b1"), SimpleNamespace(name="b2")
    t = Track("1", [b1, b2], 1, name="test123")
    # then
    with pytest.raises(ValueError):
        _ = t.band


@pytest.mark.parametrize("attr, val, expected", [("country", SimpleNamespace(text="Poland"), Countries.POLAND),
                                                 ("status", SimpleNamespace(text=2), 2),
                                                 ("label", SimpleNamespace(text=3), 3)])
def test_Band_properties(mocker, attr, val, expected):
    # given
    mocker.patch("enmet.pages._DataPage._get_header_item", lambda p1, p2: val)
    # when
    b = Band("dummy")
    # then
    assert getattr(b, attr) == expected


def test_Album_properties_reviews(mocker):
    # given
    class Dummy:
        text = "text123"

        def select_one(self, _):
            return None
    mocker.patch("enmet.pages._DataPage._get_header_item", lambda p1, p2: Dummy())
    # when
    b = Album("dummy").reviews
    # then
    assert b == (None, "text123")


def test_track_split_name_without_band():
    # given
    b1, b2 = SimpleNamespace(name="b1"), SimpleNamespace(name="b2")
    t = Track("123", [b1, b2], 1, "name1")
    # when
    name = t.name
    # then
    assert name == "name1"


def test_track_get_cached_instance():
    # given
    t1 = Track("123", [1, 2], 1, "")
    t2 = Track("123", [1, 2], 1, "")
    # then
    assert t1 is t2


def test_album_get_year_when_no_year_in_init():
    # given
    a = Album("1")
    setattr(a, "release_date", PartialDate(year=1999))
    # when
    y = a.year
    # then
    assert y == 1999


def test_EnmetEntity_re_init():
    # given
    class Test(EnmetEntity):
        pass
    # when
    obj1 = Test("1")
    obj2 = Test("1")
    # then
    assert obj1 is obj2


def test_create_default_cache(mocker):
    #given
    cs_mock = mocker.patch("enmet.pages.CachedSession")
    cs_mock.return_value.get.return_value = SimpleNamespace(content="<HTML />", raise_for_status=lambda: None)
    cp_mock = mocker.patch("enmet.pages._CachedSite._CACHE_PATH")

    class Dummy:
        RESOURCE = "ABC_resource"
        id = "123890"
        e = _CachedSite()
    # when
    result = Dummy().e
    # then
    assert result == BeautifulSoup("<html />", features="html.parser")
    assert cp_mock.method_calls == [call.mkdir(parents=True, exist_ok=True)]
    assert call(cache_name=ANY, backend="sqlite") in cs_mock.mock_calls



from typing import List

from enmet import Countries, country_to_enum_name
from enmet.common import ReleaseTypes, url_to_id, datestr_to_date
from enmet.entities import Band, Album
from enmet.pages import BandSearchPage, AlbumSearchPage, RandomBandPage

__all__ = ["search_albums", "search_bands", "random_band"]

_RELEASE_TYPE_IDS = {ReleaseTypes.FULL: 1, ReleaseTypes.LIVE: 2, ReleaseTypes.DEMO: 3, ReleaseTypes.SINGLE: 4,
                     ReleaseTypes.EP: 5, ReleaseTypes.VIDEO: 6, ReleaseTypes.BOX: 7, ReleaseTypes.SPLIT: 8,
                     ReleaseTypes.COMPILATION: 10, ReleaseTypes.SPLIT_VIDEO: 12, ReleaseTypes.COLLABORATION: 13}


_BAND_SEARCH_FIELDS_MAPPING = {
    "name": "bandName",
    "strict": "exactBandMatch",
    "genre": "genre",
    "countries": "country[]",
    "formed_from": "yearCreationFrom",
    "formed_to": "yearCreationTo",
}


def search_bands(*, name: str = None, strict: bool = None, genre: str = None, countries: List[Countries] = None,
                 formed_from: int = None, formed_to: int = None) -> List[Band]:
    if not any(locals().values()):
        return []
    params = {_BAND_SEARCH_FIELDS_MAPPING[k]: v or "" for k, v in locals().items()}
    params[_BAND_SEARCH_FIELDS_MAPPING["countries"]] = [c.value for c in countries or []]
    return [Band(url_to_id(b[0]),
                 name=b[1],
                 country=countries[0] if countries and len(countries) == 1 else Countries[country_to_enum_name(b[3])])
            for b in BandSearchPage(params).bands]


_ALBUM_SEARCH_FIELDS_MAPPING = {
    "name": "releaseTitle",
    "strict": "exactReleaseMatch",
    "band": "bandName",
    "band_strict": "exactBandMatch",
    "year_from": "releaseYearFrom",
    "month_from": "releaseMonthFrom",
    "year_to": "releaseYearTo",
    "month_to": "releaseMonthTo",
    "genre": "genre",
    "release_types": "releaseType[]"
}


def search_albums(*, name: str = None, strict: bool = None, band: str = None, band_strict: bool = None,
                  year_from: int = None, month_from: int = None, year_to: int = None, month_to: int = None,
                  genre: str = None, release_types: List[ReleaseTypes] = None):
    if not any(locals().values()):
        return []
    params = {_ALBUM_SEARCH_FIELDS_MAPPING[k]: v or "" for k, v in locals().items()}
    params[_ALBUM_SEARCH_FIELDS_MAPPING["release_types"]] = [_RELEASE_TYPE_IDS[rt] for rt in release_types or []]
    # Year is forced so that it is included in search results
    if year_from is None:
        params[_ALBUM_SEARCH_FIELDS_MAPPING["year_from"]] = 1900
    if year_to is None:
        params[_ALBUM_SEARCH_FIELDS_MAPPING["year_to"]] = 2999
    return [Album(url_to_id(a[0]), name=a[1], year=datestr_to_date(a[4]).year)
            for a
            in AlbumSearchPage(params).albums]


def random_band() -> Band:
    """Just get a random band."""
    return Band(url_to_id(RandomBandPage().band))



import re
from abc import ABC
from datetime import datetime, timedelta
from functools import cached_property, reduce
from inspect import getmembers
from typing import List, Iterable, Optional, Tuple, Union

from .countries import Countries, country_to_enum_name
from .common import CachedInstance, ReleaseTypes, url_to_id, datestr_to_date, PartialDate
from .pages import BandPage, DiscographyPage, BandRecommendationsPage, AlbumPage, LyricsPage, ArtistPage

__all__ = ["ReleaseTypes", "Entity", "ExternalEntity", "EnmetEntity",
           "DynamicEnmetEntity", "Band", "Album", "Disc", "Track", "Artist", "EntityArtist", "LineupArtist",
           "AlbumArtist", "SimilarBand"]


def _timestr_to_time(time_string: str) -> Optional[timedelta]:
    """Convert time presented as [hh:mm|mm]:ss into timedelta."""
    if not time_string:
        return None
    data = dict(zip(["seconds", "minutes", "hours"], reversed([int(t) for t in time_string.split(":")])))
    return timedelta(**data)


def _discstr_to_name(name: Optional[str]) -> Optional[str]:
    """Determine disc name (if any) from header item of track list (like "Disc 1 - Gloom")."""
    if name and "- " in name:
        return name[name.index("-")+2:].strip()
    else:
        return None


def _turn_na_into_none(data: Union[str, List, timedelta]) -> Union[List, None, str]:
    if isinstance(data, list) and len(data) == 1 and data[0].lower() == "n/a":
        return []
    elif isinstance(data, timedelta) and data == timedelta(0):
        return None
    elif isinstance(data, str) and data.lower() in ["n/a", "unknown"] or data == "":
        return None
    else:
        return data


class Entity(ABC):
    """A thing, like band or album"""
    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.name}>"

    def __dir__(self) -> List[str]:
        return [p[0] for p in getmembers(self.__class__) if type(p[1]) is cached_property]


class ExternalEntity(Entity):
    """
    Non EM entity, like non-metal musician in metal album lineup.
    It has only string representation and is a class just for the
    sake of consistency.
    """
    def __init__(self, name: str):
        self.name = name

    def __dir__(self) -> Iterable[str]:
        return ["name"]


class EnmetEntity(Entity, CachedInstance, ABC):
    def __init__(self, id_):
        id_ = str(id_)
        if not hasattr(self, "id"):
            self.id = id_

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.name} ({self.id})>"


class DynamicEnmetEntity(Entity, ABC):
    """Represents entities without its own identity in Enmet, for example disc of an album"""
    def __str__(self):
        return self.name


class Band(EnmetEntity):
    """Band or artist performing as a band."""
    def __init__(self, id_: str, *, name: str = None, country: Countries = None, genres: str = None):
        if not hasattr(self, "id"):
            super().__init__(id_)
            if name is not None:
                setattr(self, "name", name)
            if country is not None:
                setattr(self, "country", country)
            if genres is not None:
                setattr(self, "genres", genres)
            self._band_page = BandPage(self.id)
            self._albums_page = DiscographyPage(self.id)

    def __str__(self):
        return f"{self.name} ({self.country})"

    @cached_property
    def name(self) -> str:
        return self._band_page.name

    @cached_property
    def country(self) -> Countries:
        return Countries[country_to_enum_name(self._band_page.country)]

    @cached_property
    def location(self) -> str:
        return _turn_na_into_none(self._band_page.location)

    @cached_property
    def formed_in(self) -> Optional[int]:
        data = self._band_page.formed_in
        value = _turn_na_into_none(data)
        return int(value) if value else value

    @cached_property
    def years_active(self) -> List[str]:
        return _turn_na_into_none(self._band_page.years_active)

    @cached_property
    def genres(self) -> List[str]:
        return self._band_page.genres

    @cached_property
    def status(self) -> Optional[str]:
        return self._band_page.status

    @cached_property
    def lyrical_themes(self) -> Optional[List[str]]:
        return _turn_na_into_none(self._band_page.lyrical_themes)

    @cached_property
    def label(self) -> str:
        return self._band_page.current_label or self._band_page.last_label

    @cached_property
    def lineup(self) -> List["LineupArtist"]:
        data = self._band_page.lineup
        return [LineupArtist(url_to_id(a[0]), self.id, a[1], a[2]) for a in data]

    @cached_property
    def past_members(self) -> List["LineupArtist"]:
        data = self._band_page.past_members
        return [LineupArtist(url_to_id(a[0]), self.id, a[1], a[2]) for a in data]

    @cached_property
    def live_musicians(self) -> List["LineupArtist"]:
        data = self._band_page.live_musicians
        return [LineupArtist(url_to_id(a[0]), self.id, a[1], a[2]) for a in data]

    @cached_property
    def discography(self) -> List["Album"]:
        """List of band's albums in chronological order."""
        return [Album(url_to_id(a[0]), name=a[1], year=a[3]) for a in self._albums_page.albums]

    @cached_property
    def similar_artists(self) -> List["SimilarBand"]:
        return [SimilarBand(url_to_id(sa[0]), self.id, sa[4], name=sa[1], country=sa[2], genres=sa[3])
                for sa in BandRecommendationsPage(self.id).similar_artists]

    @cached_property
    def info(self) -> str:
        return _turn_na_into_none(self._band_page.info)

    @cached_property
    def last_modified(self) -> datetime:
        data = self._band_page.last_modified
        year, month, day, hour, minute, second = re.search(r"(\d{4})-(\d{2})-(\d{2})\s+(\d{2}):(\d{2}):(\d{2})",
                                                           data).groups()
        return datetime(year=int(year), month=int(month), day=int(day), hour=int(hour), minute=int(minute),
                        second=int(second))


class SimilarBand(DynamicEnmetEntity):
    def __init__(self, id_: str, similar_to_id: str, score: str, name: str = None, country: str = None,
                 genres: str = None):
        self.band = Band(id_, name=name, country=country, genres=genres)
        self.similar_to = Band(similar_to_id)
        self.score = int(score)

    def __dir__(self) -> List[str]:
        return dir(self.band) + ["score", "similar_to"]

    def __getattr__(self, item):
        return getattr(self.band, item)

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.band.name} ({self.score})>"


class Album(EnmetEntity):
    def __init__(self, id_: str, *, name: str = None, year: int = None):
        # Have parameters for str and repr ready
        if not hasattr(self, "id"):
            super().__init__(id_)
            if name is not None:
                setattr(self, "name", name)
            if year is not None:
                setattr(self, "year", year)
            self._album_page = AlbumPage(self.id)

    def __repr__(self):
        return f"<Album: {self.name} ({self.id})>"

    def __str__(self):
        return f"{self.name} ({self.year})"

    @cached_property
    def name(self) -> str:
        return self._album_page.name

    @cached_property
    def bands(self) -> List[Band]:
        return [Band(url_to_id(bid[0])) for bid in self._album_page.bands]

    @cached_property
    def type(self) -> ReleaseTypes:
        return ReleaseTypes(self._album_page.type)

    @cached_property
    def year(self) -> int:
        return self.release_date.year

    @cached_property
    def release_date(self) -> PartialDate:
        return datestr_to_date(self._album_page.release_date)

    @cached_property
    def label(self) -> str:
        return self._album_page.label

    @cached_property
    def format(self) -> str:
        return _turn_na_into_none(self._album_page.format)

    @cached_property
    def reviews(self) -> Tuple[str, str]:
        return self._album_page.reviews

    @cached_property
    def catalog_id(self) -> str:
        return _turn_na_into_none(self._album_page.catalog_id)

    @cached_property
    def discs(self) -> List["Disc"]:
        return [Disc(self.id, idx, self.bands) for idx in range(len(self._album_page.disc_names))]

    @cached_property
    def lineup(self) -> List["AlbumArtist"]:
        return [AlbumArtist(url_to_id(a[0]), self.id, name=a[1], role=a[2]) for a in self._album_page.lineup]

    @cached_property
    def total_time(self) -> timedelta:
        return _turn_na_into_none(reduce(timedelta.__add__, [disc.total_time for disc in self.discs if disc.total_time], timedelta()))


class Disc(DynamicEnmetEntity):
    def __init__(self, album_id: str, number: int = 0, bands: List[Band] = None):
        self._number = number
        self._album_page = AlbumPage(album_id)
        self._bands = bands

    @cached_property
    def number(self) -> int:
        return self._number + 1

    @cached_property
    def name(self) -> Optional[str]:
        return _discstr_to_name(self._album_page.disc_names[self._number])

    @cached_property
    def total_time(self) -> timedelta:
        return _timestr_to_time(self._album_page.total_times[self._number])

    @cached_property
    def tracks(self) -> List["Track"]:
        tracks = []
        for t in self._album_page.tracks[self._number]:
            tracks.append(Track(t[0], self._bands, t[1], t[2], _timestr_to_time(t[3]), t[4]))
        return tracks


class Track(EnmetEntity):
    def __init__(self, id_: str, bands: List[Band], number: int, name: str, time: timedelta = None,
                 lyrics_info: Optional[bool] = None):
        if not hasattr(self, "id"):
            super().__init__(id_)
            self.number = number
            self.time = time
            self._name = name
            self._lyrics_info = lyrics_info
            self._bands = bands

    def __dir__(self) -> List[str]:
        return super().__dir__() + ["number", "time"]

    @cached_property
    def name(self) -> str:
        if len(self._bands) < 2:
            return self._name
        else:
            for b in self._bands:  # Handle track on a split album
                if self._name.startswith(b.name):
                    return self._name[self._name.index("-")+2:]
            else:
                return self._name  # Probably wrong band name put into track name

    @cached_property
    def band(self) -> Band:
        if len(self._bands) == 1:
            return self._bands[0]
        else:
            for b in self._bands:
                if self._name.startswith(b.name):
                    return b
            else:
                raise ValueError("No band available for split album track")

    @cached_property
    def lyrics(self) -> Optional[Union[bool, str]]:
        if self._lyrics_info is None:
            return None  # No information
        elif self._lyrics_info is False:  # Instrumental
            return False
        else:
            return LyricsPage(self.id).lyrics


class Artist(EnmetEntity):
    """General artist info"""

    def __init__(self, id_):
        if not hasattr(self, "id"):
            super().__init__(id_)
            self._artist_page = ArtistPage(id_)

    def __str__(self):
        return f"{self.name}"

    def __dir__(self) -> Iterable[str]:
        return [p[0] for p in getmembers(self.__class__) if type(p[1]) is cached_property]

    @cached_property
    def name(self) -> str:
        return self._artist_page.name

    @cached_property
    def real_full_name(self) -> str:
        return self._artist_page.real_full_name

    @cached_property
    def age(self) -> str:
        return self._artist_page.age

    @cached_property
    def place_of_birth(self) -> str:
        return self._artist_page.place_of_birth

    @cached_property
    def gender(self) -> str:
        return self._artist_page.gender

    @cached_property
    def biography(self) -> str:
        return self._artist_page.biography

    @cached_property
    def trivia(self) -> str:
        return self._artist_page.trivia


class EntityArtist(DynamicEnmetEntity, ABC):
    """"Album artist or lineup artist"""

    def __init__(self, id_, role: str = None):
        self.artist = Artist(id_)
        self.role = role

    def __getattr__(self, item):
        return getattr(self.artist, item)

    def __dir__(self) -> List[str]:
        return dir(self.artist) + ["role"]


class LineupArtist(EntityArtist):
    """Artist in the current band lineup"""

    def __init__(self, id_: str, band_id: str, name=None, role=None):
        super().__init__(id_, role)
        self.name_in_lineup = name
        self.band = Band(band_id)

    def __dir__(self) -> Iterable[str]:
        return super().__dir__() + ["name_in_lineup", "band"]

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.name_in_lineup} ({self.id})>"

    def __str__(self):
        return self.name_in_lineup


class AlbumArtist(EntityArtist):
    """Artist for an album"""

    def __init__(self, id_: str, album_id: str, *, name: str = None, role: str = None):
        super().__init__(id_, role)
        self.name_on_album = name
        self.album = Album(album_id)

    def __dir__(self) -> Iterable[str]:
        return super().__dir__() + ["name_on_album", "album"]

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.name_on_album} ({self.id})>"

    def __str__(self):
        return self.name_on_album



"""Country list right from Metal Archives, as its country list does not match exactly any standard."""
from enum import Enum

from bs4 import BeautifulSoup

__all__ = ["Countries", "country_to_enum_name"]

# Copied from search page
DATA = '''
<select name="country"
          id="country"
          class="countrySelect">
    <option value="">(Any)</option>
    <option value="AF">Afghanistan</option>
    <option value="AX">Åland Islands</option>
    <option value="AL">Albania</option>
    <option value="DZ">Algeria</option>
    <option value="AS">American Samoa</option>
    <option value="AD">Andorra</option>
    <option value="AO">Angola</option>
    <option value="AI">Anguilla</option>
    <option value="AQ">Antarctica</option>
    <option value="AG">Antigua and Barbuda</option>
    <option value="AR">Argentina</option>
    <option value="AM">Armenia</option>
    <option value="AW">Aruba</option>
    <option value="AU">Australia</option>
    <option value="AT">Austria</option>
    <option value="AZ">Azerbaijan</option>
    <option value="BS">Bahamas</option>
    <option value="BH">Bahrain</option>
    <option value="BD">Bangladesh</option>
    <option value="BB">Barbados</option>
    <option value="BY">Belarus</option>
    <option value="BE">Belgium</option>
    <option value="BZ">Belize</option>
    <option value="BJ">Benin</option>
    <option value="BM">Bermuda</option>
    <option value="BT">Bhutan</option>
    <option value="BO">Bolivia</option>
    <option value="BQ">Bonaire, Sint Eustatius and Saba</option>
    <option value="BA">Bosnia and Herzegovina</option>
    <option value="BW">Botswana</option>
    <option value="BV">Bouvet Island</option>
    <option value="BR">Brazil</option>
    <option value="IO">British Indian Ocean Territory</option>
    <option value="BN">Brunei</option>
    <option value="BG">Bulgaria</option>
    <option value="BF">Burkina Faso</option>
    <option value="BI">Burundi</option>
    <option value="KH">Cambodia</option>
    <option value="CM">Cameroon</option>
    <option value="CA">Canada</option>
    <option value="CV">Cape Verde</option>
    <option value="KY">Cayman Islands</option>
    <option value="CF">Central African Republic</option>
    <option value="TD">Chad</option>
    <option value="CL">Chile</option>
    <option value="CN">China</option>
    <option value="CX">Christmas Island</option>
    <option value="CC">Cocos (Keeling) Islands</option>
    <option value="CO">Colombia</option>
    <option value="KM">Comoros</option>
    <option value="CD">Congo, Democratic Republic of</option>
    <option value="CG">Congo, Republic of</option>
    <option value="CK">Cook Islands</option>
    <option value="CR">Costa Rica</option>
    <option value="HR">Croatia</option>
    <option value="CU">Cuba</option>
    <option value="CW">Curaçao</option>
    <option value="CY">Cyprus</option>
    <option value="CZ">Czechia</option>
    <option value="DK">Denmark</option>
    <option value="DJ">Djibouti</option>
    <option value="DM">Dominica</option>
    <option value="DO">Dominican Republic</option>
    <option value="TL">East Timor</option>
    <option value="EC">Ecuador</option>
    <option value="EG">Egypt</option>
    <option value="SV">El Salvador</option>
    <option value="GQ">Equatorial Guinea</option>
    <option value="ER">Eritrea</option>
    <option value="EE">Estonia</option>
    <option value="SZ">Eswatini</option>
    <option value="ET">Ethiopia</option>
    <option value="FK">Falkland Islands</option>
    <option value="FO">Faroe Islands</option>
    <option value="FJ">Fiji</option>
    <option value="FI">Finland</option>
    <option value="FR">France</option>
    <option value="GF">French Guiana</option>
    <option value="PF">French Polynesia</option>
    <option value="TF">French Southern Territories</option>
    <option value="GA">Gabon</option>
    <option value="GM">Gambia</option>
    <option value="GE">Georgia</option>
    <option value="DE">Germany</option>
    <option value="GH">Ghana</option>
    <option value="GI">Gibraltar</option>
    <option value="GR">Greece</option>
    <option value="GL">Greenland</option>
    <option value="GD">Grenada</option>
    <option value="GP">Guadeloupe</option>
    <option value="GU">Guam</option>
    <option value="GT">Guatemala</option>
    <option value="GG">Guernsey</option>
    <option value="GN">Guinea</option>
    <option value="GW">Guinea-Bissau</option>
    <option value="GY">Guyana</option>
    <option value="HT">Haiti</option>
    <option value="HM">Heard and McDonald Islands</option>
    <option value="HN">Honduras</option>
    <option value="HK">Hong Kong</option>
    <option value="HU">Hungary</option>
    <option value="IS">Iceland</option>
    <option value="IN">India</option>
    <option value="ID">Indonesia</option>
    <option value="XX">International</option>
    <option value="IR">Iran</option>
    <option value="IQ">Iraq</option>
    <option value="IE">Ireland</option>
    <option value="IM">Isle of Man</option>
    <option value="IL">Israel</option>
    <option value="IT">Italy</option>
    <option value="CI">Ivory Coast</option>
    <option value="JM">Jamaica</option>
    <option value="JP">Japan</option>
    <option value="JE">Jersey</option>
    <option value="JO">Jordan</option>
    <option value="KZ">Kazakhstan</option>
    <option value="KE">Kenya</option>
    <option value="KI">Kiribati</option>
    <option value="KP">Korea, North</option>
    <option value="KR">Korea, South</option>
    <option value="KW">Kuwait</option>
    <option value="KG">Kyrgyzstan</option>
    <option value="LA">Laos</option>
    <option value="LV">Latvia</option>
    <option value="LB">Lebanon</option>
    <option value="LS">Lesotho</option>
    <option value="LR">Liberia</option>
    <option value="LY">Libya</option>
    <option value="LI">Liechtenstein</option>
    <option value="LT">Lithuania</option>
    <option value="LU">Luxembourg</option>
    <option value="MO">Macau</option>
    <option value="MG">Madagascar</option>
    <option value="MW">Malawi</option>
    <option value="MY">Malaysia</option>
    <option value="MV">Maldives</option>
    <option value="ML">Mali</option>
    <option value="MT">Malta</option>
    <option value="MH">Marshall Islands</option>
    <option value="MQ">Martinique</option>
    <option value="MR">Mauritania</option>
    <option value="MU">Mauritius</option>
    <option value="YT">Mayotte</option>
    <option value="MX">Mexico</option>
    <option value="FM">Micronesia, Federated States of</option>
    <option value="MD">Moldova</option>
    <option value="MC">Monaco</option>
    <option value="MN">Mongolia</option>
    <option value="ME">Montenegro</option>
    <option value="MS">Montserrat</option>
    <option value="MA">Morocco</option>
    <option value="MZ">Mozambique</option>
    <option value="MM">Myanmar</option>
    <option value="NA">Namibia</option>
    <option value="NR">Nauru</option>
    <option value="NP">Nepal</option>
    <option value="NL">Netherlands</option>
    <option value="NC">New Caledonia</option>
    <option value="NZ">New Zealand</option>
    <option value="NI">Nicaragua</option>
    <option value="NE">Niger</option>
    <option value="NG">Nigeria</option>
    <option value="NU">Niue</option>
    <option value="NF">Norfolk Island</option>
    <option value="MK">North Macedonia</option>
    <option value="MP">Northern Mariana Islands</option>
    <option value="NO">Norway</option>
    <option value="OM">Oman</option>
    <option value="PK">Pakistan</option>
    <option value="PW">Palau</option>
    <option value="PS">Palestine</option>
    <option value="PA">Panama</option>
    <option value="PG">Papua New Guinea</option>
    <option value="PY">Paraguay</option>
    <option value="PE">Peru</option>
    <option value="PH">Philippines</option>
    <option value="PN">Pitcairn Island</option>
    <option value="PL">Poland</option>
    <option value="PT">Portugal</option>
    <option value="PR">Puerto Rico</option>
    <option value="QA">Qatar</option>
    <option value="RE">Reunion</option>
    <option value="RO">Romania</option>
    <option value="RU">Russia</option>
    <option value="RW">Rwanda</option>
    <option value="BL">Saint Barthélemy</option>
    <option value="KN">Saint Kitts &amp; Nevis</option>
    <option value="LC">Saint Lucia</option>
    <option value="MF">Saint Martin (French part)</option>
    <option value="PM">Saint Pierre and Miquelon</option>
    <option value="VC">Saint Vincent and The Grenadines</option>
    <option value="WS">Samoa</option>
    <option value="SM">San Marino</option>
    <option value="ST">Sao Tome and Principe</option>
    <option value="SA">Saudi Arabia</option>
    <option value="SN">Senegal</option>
    <option value="RS">Serbia</option>
    <option value="SC">Seychelles</option>
    <option value="SL">Sierra Leone</option>
    <option value="SG">Singapore</option>
    <option value="SX">Sint Maarten (Dutch part)</option>
    <option value="SK">Slovakia</option>
    <option value="SI">Slovenia</option>
    <option value="SB">Solomon Islands</option>
    <option value="SO">Somalia</option>
    <option value="ZA">South Africa</option>
    <option value="GS">South Georgia &amp; South Sandwich Islands</option>
    <option value="SS">South Sudan</option>
    <option value="ES">Spain</option>
    <option value="LK">Sri Lanka</option>
    <option value="SH">St Helena, Ascension &amp; Tristan da Cunha</option>
    <option value="SD">Sudan</option>
    <option value="SR">Suriname</option>
    <option value="SJ">Svalbard</option>
    <option value="SE">Sweden</option>
    <option value="CH">Switzerland</option>
    <option value="SY">Syria</option>
    <option value="TW">Taiwan</option>
    <option value="TJ">Tajikistan</option>
    <option value="TZ">Tanzania</option>
    <option value="TH">Thailand</option>
    <option value="TG">Togo</option>
    <option value="TK">Tokelau</option>
    <option value="TO">Tonga</option>
    <option value="TT">Trinidad and Tobago</option>
    <option value="TN">Tunisia</option>
    <option value="TR">Türkiye</option>
    <option value="TM">Turkmenistan</option>
    <option value="TC">Turks and Caicos Islands</option>
    <option value="TV">Tuvalu</option>
    <option value="UM">U.S. Minor Outlying Islands</option>
    <option value="UG">Uganda</option>
    <option value="UA">Ukraine</option>
    <option value="AE">United Arab Emirates</option>
    <option value="GB">United Kingdom</option>
    <option value="US">United States</option>
    <option value="ZZ">Unknown</option>
    <option value="UY">Uruguay</option>
    <option value="UZ">Uzbekistan</option>
    <option value="VU">Vanuatu</option>
    <option value="VA">Vatican City</option>
    <option value="VE">Venezuela</option>
    <option value="VN">Vietnam</option>
    <option value="VG">Virgin Islands (British)</option>
    <option value="VI">Virgin Islands (US)</option>
    <option value="WF">Wallis and Futuna</option>
    <option value="EH">Western Sahara</option>
    <option value="YE">Yemen</option>
    <option value="ZM">Zambia</option>
    <option value="ZW">Zimbabwe</option>
</select>
'''


def country_to_enum_name(country: str) -> str:
    return "".join(filter(lambda x: x.isalpha() or x == "_",
                          country.replace(" ", "_").
                          replace("ü", "u").
                          replace("é", "e").
                          replace("ç", "c").
                          replace("Å", "A").upper()))


def _enum_str(self):
    return self.name.replace('_', ' ').title()


Countries = Enum("Countries",
                 {country_to_enum_name(elem.text): elem["value"]
                  for elem
                  in BeautifulSoup(DATA, features="html.parser").find_all("option")
                  if elem["value"]
                  }
                 |
                 {"__str__": _enum_str}
                 )



import logging
from enum import Enum
from pathlib import PurePath
from urllib.parse import urlparse
from weakref import WeakValueDictionary

__all__ = ["ReleaseTypes", "PartialDate"]

_logger = logging.getLogger(__file__)


def url_to_id(url: str) -> str:
    """Extract id from URL, if id is the last part of path: http://host.com/path/more_path/id?param=value."""
    _, _, path, *_ = urlparse(url)
    return PurePath(path).name


class PartialDate:
    """Keeps date that has only year, only year and month or year, month and day components."""
    DAYS_IN_MONTH = [0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    MONTHS = ["", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October",
              "November", "December"]

    def __init__(self, year: int, month: str = None, day: int = None):
        """ Supports pre-parsed format used on Metal Archives pages: '14th September 1984'"""
        days_in_month = 99  # Just to sedate linters
        if month and month not in self.MONTHS[1:]:
            raise ValueError(f"Invalid month value: {month}")
        if month is not None:
            month = self.MONTHS.index(month)
            days_in_month = self.DAYS_IN_MONTH[month]
            days_in_month += 1 if month == 2 and year % 4 == 0 and (year % 100 != 0 or year % 400 == 0) else 0
        if any([day and not month,
                day and not (1 <= day <= days_in_month)]):
            raise ValueError(f"Invalid date values: {year}, {month}, {day}")
        self.year = year
        self.month = month
        self.day = day

    def __repr__(self):
        return f"<{self.__class__.__name__}: year={self.year}, month={self.month}, day={self.day}>"

    def __str__(self):
        return f"{self.year}{'' if self.month is None else '-{:02}'.format(self.month)}" \
               f"{'' if self.day is None else '-{:02}'.format(self.day)}"

    def __eq__(self, other):
        return self.year == other.year and self.month == other.month and self.day == other.day


def datestr_to_date(date_string: str) -> PartialDate:
    """Convert date string as used on Metal Archives pages ('14th September 1984') into PartialDate object."""
    match date_string.split():
        case month, day, year:  # February 19th, 1981
            day = "".join(filter(str.isdigit, day))
            return PartialDate(day=int(day), month=month, year=int(year))
        case month, year:  # September 1981
            return PartialDate(month=month, year=int(year))
        case year:  # 1981
            return PartialDate(year=int(year[0]))


class CachedInstance:
    """Mixin to reuse existing objects."""
    _CACHE = WeakValueDictionary()

    def __new__(cls, id_: str, *args, **kwargs):
        id_ = str(id_)
        if obj := CachedInstance._CACHE.get((cls.__name__, id_)):
            _logger.debug(f"cached get {cls.__name__} {id_}")
            return obj
        else:
            _logger.debug(f"uncached get {cls.__name__} {id_}")
            obj = super().__new__(cls)
            CachedInstance._CACHE[(cls.__name__, id_)] = obj
            return obj


class ReleaseTypes(Enum):
    """Names for release types."""
    FULL = "Full-length"
    LIVE = "Live album"
    DEMO = "Demo"
    SINGLE = "Single"
    EP = "EP"
    VIDEO = "Video"
    BOX = "Boxed set"
    SPLIT = "Split"
    COMPILATION = "Compilation"
    SPLIT_VIDEO = "Split video"
    COLLABORATION = "Collaboration"
